#include "reader.hpp"
#include <stdlib.h>
#include <string.h>
#include "math_utils.hpp"

using namespace Voltmeter;


Reader::Reader(Player & player)
    : player(&player)
    , playlistLength(0)
    , steadyReadings(0)
    , didReadSteadyVoltage(false)
{}


void Reader::Execute()
{
    if (!player->IsPlaying() && playlistLength > 0)
    {
        char fileName[16] = "sounds/";
        int fileIndex = (int) playlist[0];
        itoa(fileIndex, fileName + 7, 10);
        strcat(fileName, ".wav");
        
        if (!player->OpenFile(fileName))
        {
            player->OpenFile("sounds/254.wav");
        }

        --playlistLength;

        for (int8_t i = 0; i < playlistLength; ++i)
        {
            playlist[i] = playlist[i + 1];
        }
    }
}


bool Reader::IsReading()
{
    return ((playlistLength > 0) || player->IsPlaying());
}


void Reader::StopReading(bool finishCurrentFile)
{
    playlistLength = 0;

    if (!finishCurrentFile)
    {
        // todo: player.Stop();
    }
}



void Reader::ReadMeasurement(Voltage const & newAcVoltage, Voltage const & newDcVoltage)
{
    if (newAcVoltage.IsEqualTo(steadyAcVoltage, 2) && newDcVoltage.IsEqualTo(steadyDcVoltage, 2))
    {
        ++steadyReadings;
    }
    else
    {
        steadyReadings = 0;
        steadyAcVoltage = newAcVoltage;
        steadyDcVoltage = newDcVoltage;
        didReadSteadyVoltage = false;
    }

    if ((steadyReadings >= 4) && !didReadSteadyVoltage && !IsReading())
    {
        if (!newAcVoltage.IsEqualTo(lastReadAcVoltage, 2) || !newDcVoltage.IsEqualTo(lastReadDcVoltage, 2))
        {
            this->ForceReadMeasurement(newAcVoltage, newDcVoltage);
        }

        didReadSteadyVoltage = true;
    }
}


void Reader::ForceReadMeasurement(Voltage const & acVoltage, Voltage const & dcVoltage)
{
    this->StopReading(false);

    bool readAc = MathUtils::Abs(acVoltage.value) > 2;

    if (!readAc || (MathUtils::Abs(dcVoltage.value) > 2))
    {
        this->ReadVoltage(false, dcVoltage);
    }

    if (readAc)
    {
        this->ReadVoltage(true, acVoltage);
    }


    this->PlayEndReadingSound();

    this->lastReadAcVoltage = acVoltage;
    this->lastReadDcVoltage = dcVoltage;
}


void Reader::ReadRange(Range const range)
{
    this->StopReading(false);
    
    switch (range)
    {
        case Range::_400mV:
            this->AddToPlaylist(File::Zakres400mV);
            break;

        case Range::_4V:
            this->AddToPlaylist(File::Zakres4V);
            break;

        case Range::_40V:
            this->AddToPlaylist(File::Zakres40V);
            break;

        case Range::_400V:
            this->AddToPlaylist(File::Zakres400V);
            break;
    }

    this->PlayEndReadingSound();
}


void Reader::ReadAutoRange()
{
    this->StopReading(false);
    this->AddToPlaylist(File::ZakresAutomatyczny);
    this->PlayEndReadingSound();
}


void Reader::ReadCalibration()
{
    this->StopReading(true);
    this->AddToPlaylist(File::Cisza);
    this->AddToPlaylist(File::Working);
}

void Reader::AddToPlaylist(File const file)
{
    if (playlistLength < MaxPlaylistLength)
    {
        playlist[playlistLength] = file;
        ++playlistLength;
    }
}


void Reader::ReadHundreds(int16_t number)
{
    if (number >= 100)
    {
        switch (number / 100)
        {
            case 1:
                this->AddToPlaylist(File::Sto);
                break;

            case 2:
                this->AddToPlaylist(File::Dwiescie);
                break;

            case 3:
                this->AddToPlaylist(File::Trzysta);
                break;

            default:
                this->AddToPlaylist(File::None);
                break;
        }
    } 
}


void Reader::ReadTens(int16_t number)
{
    if (number >= 20)
    {
        switch (number / 10)
        {
            case 2:
                this->AddToPlaylist(File::Dwadziescia);
                break;

            case 3:
                this->AddToPlaylist(File::Trzydziesci);
                break;

            case 4:
                this->AddToPlaylist(File::Czterdziesci);
                break;

            case 5:
                this->AddToPlaylist(File::Piecdziesiat);
                break;

            case 6:
                this->AddToPlaylist(File::Szescdziesiat);
                break;

            case 7:
                this->AddToPlaylist(File::Siedemdziesiat);
                break;

            case 8:
                this->AddToPlaylist(File::Osiemdziesiat);
                break;

            case 9:
                this->AddToPlaylist(File::Dziewiecdziesiat);
                break;

            default:
                this->AddToPlaylist(File::None);
                break;
        }
    } 
}


void Reader::ReadDigits(int16_t number, bool isGenitive)
{
    if (number == 0)
    {
        return;
    }

    if (isGenitive && (number == 2))
    {
        this->AddToPlaylist(File::Dwie);
    }
    else
    {
        this->AddToPlaylist((File) number);
    }
}


void Reader::ReadNumber(int16_t number, bool isGenitive = false)
{
    if (number == 0)
    {
        this->AddToPlaylist(File::Zero);
    }
    else
    {
        this->ReadHundreds(number);
        number = number % 100;
        this->ReadTens(number);

        if (isGenitive && (number == 1))
        {
            this->AddToPlaylist(File::Jedna);
        }
        else if (number < 20)
        {
            this->ReadDigits(number % 20, isGenitive);
        }
        else
        {
            this->ReadDigits(number % 10, isGenitive);
        }
    }
}


void Reader::NounAdjectiveForm(int16_t const number, File singular, File pluralNominative, File pluralGenitive)
{
    if (number == 1)
    {
        this->AddToPlaylist(singular);
    }
    else if ((number % 10 >= 2) && (number % 10 <= 4) && (number < 10 || number > 20))
    {
        this->AddToPlaylist(pluralNominative);
    }
    else
    {
        this->AddToPlaylist(pluralGenitive);
    }
}


void Reader::ReadFractionalPart(int16_t number, int16_t denominator)
{
    while (number % 10 == 0)
    {
        number /= 10;
        denominator /= 10;
    }

    if ((number == 5) and (denominator == 10))
    {
        this->AddToPlaylist(File::Pol);
    }
    else
    {
        if (number == 1)
        {
            this->AddToPlaylist(File::Jedna);            
        }
        else if (number == 2)
        {
            this->AddToPlaylist(File::Dwie);
        }
        else
        {
            this->ReadNumber(number, true);
        }

        if (denominator == 10)
        {
            this->NounAdjectiveForm(number, File::Dziesiata, File::Dziesiate, File::Dziesiatych);
        }
        else
        {
            this->NounAdjectiveForm(number, File::Setna, File::Setne, File::Setnych);
        }
    }
}


void Reader::ReadVoltage(bool isAc, Voltage const & voltage)
{
    int16_t value = voltage.value;

    if (isAc)
    {
        this->PlayAcVoltageSound();
    }
    else
    {
        this->PlayDcVoltageSound();
    }

    if (voltage.IsOverRange())
    {
        if (value > 0)
        {
            this->AddToPlaylist(File::NapieciePozaZakresem);
        }
        else
        {
            this->AddToPlaylist(File::NapiecieUjemnePozaZakresem);
        }

        return;
    }

    if (value < 0)
    {
        this->AddToPlaylist(File::Minus);
        value = -value;
    }

    int16_t IntPartDenominator = MathUtils::Pow10(voltage.decimalPoint + 1);
    int16_t FracPartDenominator = voltage.denominator / IntPartDenominator;
    int16_t IntPart = value / FracPartDenominator;
    int16_t FracPart = value % FracPartDenominator;

    if ((IntPart == 1) && (FracPart == FracPartDenominator / 2) && (FracPartDenominator >= 10))
    {
        this->AddToPlaylist(File::Poltora);

        if (voltage.unit == Voltage::Unit::V)
        {
            this->AddToPlaylist(File::Wolta);
        }
        else
        {
            this->AddToPlaylist(File::Miliwolta);
        }

        return;
    }

    bool readFracPart = (FracPartDenominator > 1) && (FracPart != 0);
    bool readIntPart = (IntPart != 0) || !readFracPart;

    if (readIntPart)
    {
        this->ReadNumber(IntPart);
    }

    if (readFracPart)
    {
        if (readIntPart)
        {
            this->AddToPlaylist(File::I);
        }

        this->ReadFractionalPart(FracPart, FracPartDenominator);

        if (voltage.unit == Voltage::Unit::V)
        {
            this->AddToPlaylist(File::Wolta);
        }
        else
        {
            this->AddToPlaylist(File::Miliwolta);
        }
    }
    else
    {
        if (voltage.unit == Voltage::Unit::V)
        {
            this->NounAdjectiveForm(IntPart, File::Wolt, File::Wolty, File::Woltow);
        }
        else
        {
            this->NounAdjectiveForm(IntPart, File::Miliwolt, File::Miliwolty, File::Miliwoltow);
        }
    }
}


void Reader::PlayAcVoltageSound()
{
    this->AddToPlaylist(File::AcVoltage);
    this->AddToPlaylist(File::Cisza);
}


void Reader::PlayDcVoltageSound()
{
    this->AddToPlaylist(File::DcVoltage);
    this->AddToPlaylist(File::Cisza);
}


void Reader::PlayEndReadingSound()
{
    this->AddToPlaylist(File::EndReading);
    this->AddToPlaylist(File::Cisza);
}
